#include "B4RDefines.h"
 
namespace B4R {

	//returns a B4R byte array containing date/time representing the seconds given
	ArrayByte* B4RESP8266rtc::Sec2DT(ULong sec)  {
		ArrayByte* datetime = CreateStackMemoryObject(ArrayByte);
		datetime->data = StackMemory::buffer + StackMemory::cp;
		datetime->length = 8;
		B4R::StackMemory::cp += 8;
		
		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		
		return datetime;
	}
	
	//private function
	//convert seconds into an 8 byte array containing date/time
	void B4RESP8266rtc::SecAsDateTime(ULong sec, byte* dt)  {
		ULong days = sec / 86400;
		
		ULong dow = ((days+6)%7)+1;    //day of week  1=monday, 7=sunday (never 0)

		sec = sec - (days*86400);
		int hrs = sec / 3600;
		sec = sec - (hrs*3600);
		int min = sec / 60;
		sec = sec - (min*60);

		ULong lpy = days / 1461;
		days = days - (lpy*1461);
		
		ULong ya = 0;
		ULong y = 0;
		if (days>=366)  {
			ya=1;  //the first year is a leap year
			days = days - 366;
			y = (days/365);  //remainder years
			days = days - (y*365);
		}
		y = y + ya;

		ULong m;

		days++;		//day zero becomes the 1st	
		if (days<=31)  {m=1;} else { 
		  days = days - 31;
		  if (days<=28)  {m=2;} else {
			 if ((y==0) && (days == 29))  {m=2;} else {
				if (y==0)  {
					days = days - 29;	
				} else {
					days = days - 28;
				} 
			    if (days<=31)  {m=3;} else {
				  days=days-31; 
				  if (days<=30)  {m=4;} else {
				    days=days-30;
				    if (days<=31)	 {m=5;} else {
					  days=days-31;
					  if (days<=30)  {m=6;} else {
					    days=days-30; 
						if (days<=31)  {m=7;} else {
						  days=days-31;
						  if (days<=31)  {m=8;} else {
						    days=days-31;
							if (days<=30)  {m=9;} else {
							  days=days-30;
							  if (days<=31)  {m=10;} else {
								days=days-31;
								if (days<=30)  {m=11;} else {
								  days=days-30;
								   m=12;
								}
							  }
							}
						  }
						}
					 } 
				  }
				} 
			  }
		    }
		  }  
		}
		
		dt[0] = 21;  //century
		dt[1] = y + (lpy*4) + 12;  //add 12 years back in
		dt[2] = m;
		dt[3] = days;
		dt[4] = hrs;
		dt[5] = min;
		dt[6] = sec;
		dt[7] = dow;  //day of the week
	}

	//returns the current date/time as a B4R byte array
	bool B4RESP8266rtc::ReadClock(ArrayByte* datetime)  {
		if (SetAt_millis==0)  {
			return false;  //clock has not been set
		}

		ULong sec;
		sec = (millis() - SetAt_millis)/1000;
		sec = sec + Set_DateTime_AsSeconds;

		byte dt[8];
		SecAsDateTime(sec,dt);
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}

		return true;  //success
	}
	
	//returns a seconds for the B4R byte array containing a date/time
	ULong B4RESP8266rtc::DT2Sec(ArrayByte* datetime)  {
		ULong sec;
		ULong days;
		byte m;
		byte y;
		
		int dt[8];
		for (byte i = 0;i < 8;i++) {
			 dt[i] = ((byte*)datetime->data)[i];
		}
		
		sec = dt[6] + dt[5]*60 + dt[4]*3600;  //convert time to seconds
		
		days = dt[3]-1;  //1st becomes day zero
		m = dt[2];
		y = dt[1] - 12;  //subtract 12 years (to reduce the seconds count)
		
		byte lpy = y/4;
		days = days + (lpy*1461);
		y = y - (lpy*4);
		
		if (y>0)  {
			days = days + 366;
			y--;
		} else {
			if (m>2) days++;
		}
		
		days = days + (y*365);

		days = days + (m-1)*30;	
		if (m>1) days++;
		if (m>2) days=days-2;
		if (m>3) days++;
		if (m>5) days++;
		if (m>7) days++;
		if (m>8) days++;
		if (m>10) days++;
	
		sec = sec + (days*86400);
		
		return sec;
	}
	
	void B4RESP8266rtc::SetClock(ArrayByte* datetime, 
							int offset_hours, int offset_minutes)  {

		ULong sec = DT2Sec(datetime);
		ULong ofsec = ((offset_hours*60)+offset_minutes)*60;
		sec = sec + ofsec;
		
		Set_DateTime_AsSeconds = sec;
		SetAt_millis = millis();
		
		if (tick_every_second)  {
			start_one_second_tick();
		}
	}
	
	int B4RESP8266rtc::GetUTC(ArrayByte* dt, B4RString* Host, B4RString* Url)  {
		byte xdt[8];
		int j = 0;
		WiFiClient client;
		const int httpPort = 80;

		if (WiFi.status() != WL_CONNECTED) {
			//Wifi not connected
			return 2;
		}
		
		if (!client.connect(Host->data, httpPort)) {
			//Host not found
			return 3;
		}

		client.print(String("GET ") + Url->data + " HTTP/1.1\r\n" +
					"Host: " + Host->data + "\r\n" + 
					"User-Agent: esp8266 \r\n" + 
					"Connection: close\r\n\r\n");
		delay(1000);
		
				while(client.available()) {
			String line = client.readStringUntil('\r');
			if (line=="\n") {
				String line = client.readStringUntil('\r');
				line = line + ' ';
				//convert string to Long
				char buf[line.length()];
				line.toCharArray(buf,line.length());
				ULong TimeInSeconds=atol(buf); 
				//::Serial.println(TimeInSeconds);
				byte xdt[8];
				SecAsDateTime(TimeInSeconds,xdt);
				for (byte i = 0;i < 8;i++) {
					((byte*)dt->data)[i] = xdt[i];
				}
				if (xdt[0]!=0) {
					return 0;   //successfully obtained DateTime
				} else {
					return 1;   //invalid Date/Time
				}
			}
		}

		
		while(client.available()) {
			String line = client.readStringUntil('\r');
			client.stop();
			//convert String to Long
			char buf[line.length()];
			::Serial.println(line);
			line.toCharArray(buf,line.length());
			long TimeInSeconds = atol(buf); 
			//Serial.println(result);
			return TimeInSeconds;
			}
		return 4;  //invalid response from host
	}
	long string_to_long (String number)
{
  number = number + ' ';
  char buf[number.length()];
  //Serial.println(number);
  number.toCharArray(buf,number.length());
  long result=atol(buf); 
  //Serial.println(result);
  return result;
}

	// Get 8 byte Date/Time array from a server
	int B4RESP8266rtc::GetDT(ArrayByte* dt, B4RString* Host, UInt Port)  {
		byte xdt[8];
		int j = 0;
		WiFiClient client;

		if (WiFi.status() != WL_CONNECTED) {
			//Wifi not connected
			return 2;
		}
		
		if (!client.connect(Host->data, Port)) {
			//Host not found
			return 3;
		}

		delay(1000);
		
		while(client.available()) {
			int b;
			for (byte i = 0;i < 8;i++) {
				b = client.read();
				if (b == -1)  {
					client.stop();
					return 4;   //invalid response from server
				}
				((byte*)dt->data)[i] = b;
			}
			client.stop();
			if ((xdt[0]!=0) && (xdt[7]==0)) {
				return 0;   //successfully obtained DateTime
			} else {
				return 1;   //invalid Date/Time
			}
		}
		return 4;  //invalid response from host
	}

	static union FunctionUnion fncu;
	void B4RESP8266rtc::Initialize(void (*Event) (byte idx), bool tick_every_sec) {
		fncu.TimerFunction = tick;
		this->event = Event;
	
		for (byte i = 0;i < 10;i++) {
			this->enabled[i] = 0;  //0=not active
		}
		
		tick_every_second = tick_every_sec;
	}

	void B4RESP8266rtc::start_one_second_tick() {
		counter++;
		if (counter==0)  counter=1;
		this->enabled[0] = counter;
		scheduler.add(millis() + 1000, fncu, counter, this);
	}
	
	bool B4RESP8266rtc::SetAlarm(byte Index, ArrayByte* datetime) {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (1 to 9 only)
		}
		
		ULong CurrentAsSec;
		CurrentAsSec = (millis() - SetAt_millis)/1000;
		CurrentAsSec = CurrentAsSec + Set_DateTime_AsSeconds;
		ULong AlmAsSec = DT2Sec(datetime);
		if (AlmAsSec<=CurrentAsSec)  {
			return false;   //date/time has already passed
		}
		ULong Interval = (AlmAsSec - CurrentAsSec);
		Interval = Interval * 1000;
		
		counter++;
		if (counter==0)  counter=1;
		this->enabled[Index] = counter;
		scheduler.add(millis() + Interval, fncu, counter, this);
		return true;  //success
	}
	bool B4RESP8266rtc::CancelAlarm(byte Index)  {
		if ((Index==0) || (Index>9))  {
			return false;   //index out of range (0 to 9 only)
		}
		this->enabled[Index] = 0;  //0=not active
		return true;  //success
	}
	
	//called on alarm event
	void B4RESP8266rtc::tick(byte tag, void* target) {
		B4RESP8266rtc* me = (B4RESP8266rtc*)target;

		for (byte i = 0;i < 10;i++) {
			if (me->enabled[i] == tag)  {
				me->enabled[i] = 0;  //0=not active
				if (i==0)  {
					me->start_one_second_tick();  //and the next second, and so on
				}
				me->event(i);  //raise B4R alarm event with it's "Index" (0 to 9)
			}
		}
	}

	B4RString* B4RESP8266rtc::ShortDate(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[3];  //day
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';		
		
		v =((byte*)datetime->data)[2];  //month
		v=((v-1)*3);
		String months = "JanFebMarAprMayJunJulAugSepOctNovDec";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)months[v+i];		
		}
		
		StackMemory::buffer[StackMemory::cp++] = (byte)'/';	
		
		v =((byte*)datetime->data)[1];  //year
		pm.print(v);
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short date as string
	}

		B4RString* B4RESP8266rtc::Time24(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[4];  //hours
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[6];  //seconds
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //24hr time
	}

		B4RString* B4RESP8266rtc::Time12(ArrayByte* datetime) {
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte h;
		h =((byte*)datetime->data)[4];  //hours
		byte v=h%12;
		if (v==0) v=12;
		if (v<10) pm.print(0);
		pm.print(v);

		StackMemory::buffer[StackMemory::cp++] = (byte)':';		
		
		v =((byte*)datetime->data)[5];  //minutes
		if (v<10) pm.print(0);
		pm.print(v);
		
		if (h<12) {
			StackMemory::buffer[StackMemory::cp++] = (byte)'a';			
		} else {
			StackMemory::buffer[StackMemory::cp++] = (byte)'p';			
		}
		StackMemory::buffer[StackMemory::cp++] = (byte)'m';		
		
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //12hr time
	}

	B4RString* B4RESP8266rtc::ShortDOW(ArrayByte* datetime) {
		//day of week
		PrintToMemory pm;
		B4RString* s = B4RString::PrintableToString(NULL);
		
		byte v;
		v =((byte*)datetime->data)[7];  //day of week
		v=(v*3);
		String dow = "   MonTueWedThuFriSatSun";
		for (byte i = 0;i < 3;i++) {
			StackMemory::buffer[StackMemory::cp++] = (byte)dow[v+i];		
		}
		StackMemory::buffer[StackMemory::cp++] = 0;  //end of string
		
		return s;  //short day-of-week as string
	}

	
	//returns a B4R byte array containing date/time representing the seconds given
	ArrayByte* B4RESP8266rtc::ThisTime(byte this_hour, byte this_minute, byte this_second)  {
		ULong CurrentAsSec;
		CurrentAsSec = (millis() - SetAt_millis)/1000;
		CurrentAsSec = CurrentAsSec + Set_DateTime_AsSeconds;
		
		byte dt[8];
		SecAsDateTime(CurrentAsSec,dt);
		
		byte current_hour = dt[4];
		byte current_minute = dt[5];
		byte current_second = dt[6];
		
		ulong current_time = (current_hour*60*60) + (current_minute*60) + current_second;
		ulong this_time = (this_hour*60*60) + (this_minute*60) + this_second;
		
		if (this_time <= current_time) {
			//time is after mid-night
			ULong TomorrowAsSec = CurrentAsSec + (24*60*60); 
			SecAsDateTime(TomorrowAsSec,dt);
		}

		dt[4] = this_hour;
		dt[5] = this_minute;
		dt[6] = this_second;
		
		ArrayByte* datetime = CreateStackMemoryObject(ArrayByte);
		datetime->data = StackMemory::buffer + StackMemory::cp;
		datetime->length = 8;
		B4R::StackMemory::cp += 8;
		
		for (byte i = 0;i < 8;i++) {
			((byte*)datetime->data)[i] = dt[i];
		}
		
		return datetime;
	}	
	
}
